//---------------------------------------------------------------------------
#include "stdafx.h"
#pragma hdrstop

#include "Base64.h"

#pragma package(smart_init)

/////////////////////////////////////////////////////////////////////////////////
// Base 64 Encoder/Decoder Class for C++Builder
// by Alex Egorov alxeg@yandex.ru 
// Copyright 2003
/////////////////////////////////////////////////////////////////////////////////

using namespace std;
//---------------------------------------------------------------------------
char Base64::encTable[64] = { 'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
                              'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
                              '0','1','2','3','4','5','6','7','8','9','+','/' };

char Base64::decTable[256] = {
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
                              52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
                              -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
                              15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
                              -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
                              41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };

//---------------------------------------------------------------------------
void Base64::Encode(TStream *inData, TStream* outData) {

  int len = inData->Size;
  int numPaddings = len%3? (3-(len%3)) : 0;

  vector<char> out;
  out.reserve(((len / 3) + 2) * 4);
  inData->Position = 0;
  char* inBuff = new char[len + 1];
  try {
    memset((void*)inBuff, 0x00, len+1);
    inData->Read((void*)inBuff, inData->Size);
    for (int i=0; i<len-len%3; i+=3) {
      unsigned char byte0 = inBuff[i] & 0xff;
      unsigned char byte1 = inBuff[i+1] & 0xff;
      unsigned char byte2 = inBuff[i+2]  & 0xff;
      out.push_back(encTable[(byte0>>2)&0x3f]);
      out.push_back(encTable[((byte0&0x03)<<4) | ((byte1>>4)&0x0f)]);
      out.push_back(encTable[((byte1&0x0f)<<2) | ((byte2>>6)&0x03)]);
      out.push_back(encTable[byte2&0x3f]);
    }
    // process remaining bytes
    if (numPaddings==1) { // 2 bytes remaining
      unsigned char byte0=inBuff[len-len%3];
      unsigned char byte1=inBuff[(len-len%3)+1];
      out.push_back(encTable[(byte0>>2)&0x3f]);
      out.push_back(encTable[((byte0&0x03)<<4) | ((byte1>>4)&0x0f)]);
      out.push_back(encTable[(byte1&0x0f)<<2]);
      out.push_back('=');
    }
    else
      if (numPaddings==2) { // 1 bytes remaining
        unsigned char byte0=inBuff[len-len%3];
        out.push_back(encTable[(byte0>>2)&0x3f]);
        out.push_back(encTable[(byte0&0x03)<<4]);
        out.push_back('=');
        out.push_back('=');
      }
  }
  __finally {
    delete[] inBuff;
  }
  outData->Size = 0;
  outData->Write((void*)out.begin(), out.size());
  outData->Size = out.size();
  outData->Position = 0;
}
//---------------------------------------------------------------------------
void Base64::Decode(TStream *inData, TStream* outData) {
  int len = inData->Size;

  inData->Position = len - 2; char padTest[2];
  inData->Read((void*)padTest, 2);
  int diff = (padTest[0]=='='?2:(padTest[1]=='='?1:0));

  int outSize = ((len * 3) / 4) - diff;

  vector<char> out;
  out.reserve((len * 3) / 4);
  inData->Position = 0;
  UCHAR* inBuff = new UCHAR[len + 1];
  try {
    memset((void*)inBuff, 0x00, len+1);
    inData->Read((void*)inBuff, inData->Size);

    bool  cont = true;
    int   i = 0, numBytes = 0;
    unsigned int decoded = 0;

    while (cont) {
      unsigned char curChar = inBuff[i];

      if (++i>len || curChar=='=') {
        cont = false;
        numBytes = 4;
      }
      else {
        if (isspace(curChar))
          continue;

        unsigned int thisByte = decTable[curChar];
        if (thisByte==-1)
          throw Base64Exception("Invalid character in Stream '0x"+IntToHex(curChar, 2)+"'");
        decoded |= ((thisByte << (18-6*numBytes)) & (0x3f << (18-6*numBytes)));
        numBytes++;
      }

      if (numBytes>3) {
        for (int j=0; j<3; j++) {
          unsigned char data = (decoded >> (16 - j*8)) & 0xff;
          out.push_back(data);
        }
        numBytes = 0;
        decoded = 0;
      }
    }
  }
  __finally {
    delete[] inBuff;
  }
  outData->Size = 0;
  outData->Write((void*)out.begin(), out.size());
  outData->Size = outSize;
  outData->Position = 0;
}
//---------------------------------------------------------------------------

